/*
 * Copyright (C) 2008, Alexander Kudjashev <Kudjashev@gmail.com>
 *
 * This file is subject to the terms and conditions of the GNU General Public
 * License. See the file COPYING in the main directory of this archive for
 * more details.
 *
 * Layout is based on skeletonfb.c by James Simmons and Geert Uytterhoeven
 * and arcfb.c by Jaya Kumar.
 *
 * This driver was written to be used with nokia 6100 lcd
 * (philips pcf8833 132*132@256 only).
 */

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/vmalloc.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/spi/spi.h>
#include <linux/uaccess.h>


/*****************************************************************************
          Command of LCD NOKIA6100(phillips chipset)
*****************************************************************************/
#define NOP         0x00 // nop
#define SWRESET     0x01 // software reset
#define BSTROFF     0x02 // booster voltage OFF
#define BSTRON      0x03 // booster voltage ON
#define RDDIDIF     0x04 // read display identification
#define RDDST       0x09 // read display status
#define SLEEPIN     0x10 // sleep in
#define SLEEPOUT    0x11 // sleep out
#define PTLON       0x12 // partial display mode
#define NORON       0x13 // display normal mode
#define INVOFF      0x20 // inversion OFF
#define INVON       0x21 // inversion ON
#define DALO        0x22 // all pixel OFF
#define DAL         0x23 // all pixel ON
#define SETCON      0x25 // write contrast
#define DISPOFF     0x28 // display OFF
#define DISPON      0x29 // display ON
#define CASET       0x2A // column address set
#define PASET       0x2B // page address set
#define RAMWR       0x2C // memory write
#define RGBSET      0x2D // colour set
#define PTLAR       0x30 // partial area
#define VSCRDEF     0x33 // vertical scrolling definition
#define TEOFF       0x34 // test mode
#define TEON        0x35 // test mode
#define MADCTL      0x36 // memory access control
#define SEP         0x37 // vertical scrolling start address
#define IDMOFF      0x38 // idle mode OFF
#define IDMON       0x39 // idle mode ON
#define COLMOD      0x3A // interface pixel format
#define SETVOP      0xB0 // set Vop
#define BRS         0xB4 // bottom row swap
#define TRS         0xB6 // top row swap
#define DISCTR      0xB9 // display control
#define DOR         0xBA // data order
#define TCDFE       0xBD // enable/disable DF temperature compensation
#define TCVOPE      0xBF // enable/disable Vop temp comp
#define EC          0xC0 // internal or external oscillator
#define SETMUL      0xC2 // set multiplication factor
#define TCVOPAB     0xC3 // set TCVOP slopes A and B
#define TCVOPCD     0xC4 // set TCVOP slopes c and d
#define TCDF        0xC5 // set divider frequency
#define DF8COLOR    0xC6 // set divider frequency 8-color mode
#define SETBS       0xC7 // set bias system
#define RDTEMP      0xC8 // temperature read back
#define NLI         0xC9 // n-line inversion
#define RDID1       0xDA // read ID1
#define RDID2       0xDB // read ID2
#define RDID3       0xDC // read ID3

/* lcd resolution */
#define X_RES       128
#define Y_RES       128
#define B_PP        8
#define MEM_LEN     (X_RES * Y_RES * B_PP / 8)
#define BLOCK_LEN   2048
#define REFR_TIME   (HZ / 10)

struct pcf8833_par {
	struct task_struct  *pcf8833_thread_task;
	struct spi_device   *spi;
	u8                  *buffer;
	u8                  *shadow;
	u8                  *screen;
//	struct spi_message  msg;
//	struct spi_transfer xfer;
};

static void spi_command(struct spi_device *spi, unsigned int c)
{
	unsigned int w = c & ~0x0100;

	spi_write(spi, (u8 *)&w, 2);	
}

static void spi_data(struct spi_device *spi, unsigned int d)
{
	unsigned int w = d | 0x0100;
	
	spi_write(spi, (u8 *)&w, 2);	
}

static struct fb_fix_screeninfo pcf8833_fix __devinitdata = {
	.id             = "pcf8833",
	.type           = FB_TYPE_PACKED_PIXELS,
	.visual         = FB_VISUAL_PSEUDOCOLOR,
	.xpanstep       = 0,
	.ypanstep       = 0,
	.ywrapstep      = 0,
	.line_length    = X_RES * B_PP / 8,
	.accel          = FB_ACCEL_NONE,
};

static struct fb_var_screeninfo pcf8833_var __devinitdata = {
	.xres           = X_RES,
	.yres           = Y_RES,
	.xres_virtual   = X_RES,
	.yres_virtual   = Y_RES,
	.height         = -1,
	.width          = -1,
	.vmode          = FB_VMODE_NONINTERLACED,
	.bits_per_pixel = B_PP,
	.red            = { 0, 8, 0 },
	.green          = { 0, 8, 0 },
	.blue           = { 0, 8, 0 },
	.nonstd         = 0,
};

static void spi_complete(void *param)
{

}

static int pcf8833_blank(int blank_mode, struct fb_info *info)
{
	struct pcf8833_par *par = info->par;

	switch (blank_mode) {
	case FB_BLANK_UNBLANK:
		spi_command(par->spi, DISPON);
		break;
	case FB_BLANK_NORMAL:
	case FB_BLANK_VSYNC_SUSPEND:
	case FB_BLANK_HSYNC_SUSPEND:
	case FB_BLANK_POWERDOWN:
		spi_command(par->spi, DISPOFF);
		break;
	default:
		return -EINVAL;
	}

	return 0;
}

static int pcf8833_mmap(struct fb_info *info, struct vm_area_struct *vma)
{
	unsigned long size, offset;

	size = vma->vm_end - vma->vm_start;
	offset = vma->vm_pgoff << PAGE_SHIFT;

	offset += info->fix.smem_start;
	if (remap_pfn_range(vma, vma->vm_start, offset >> PAGE_SHIFT,
			    size, vma->vm_page_prot))
		return -EAGAIN;

	printk(KERN_INFO "pcf8833: mmap framebuffer P(%lx)->V(%lx), size %lx\n",
		offset, vma->vm_start, size);
	return 0;
}

static struct fb_ops pcf8833_ops = {
	.owner          = THIS_MODULE,
	.fb_read        = fb_sys_read,
	.fb_write       = fb_sys_write,
	.fb_fillrect    = sys_fillrect,
	.fb_copyarea    = sys_copyarea,
	.fb_imageblit   = sys_imageblit,
	.fb_blank       = pcf8833_blank,
	.fb_mmap        = pcf8833_mmap,
};

static int pcf8833_thread(void *param)
{
	struct pcf8833_par *par = (struct pcf8833_par *)param;
	int i;

	set_user_nice(current, 5); 

	while(!kthread_should_stop()) {
		/* Optimize refresh: transfer image only when it has changed */
		if(memcmp(par->shadow, par->buffer, MEM_LEN)) {
			memcpy(par->shadow, par->buffer, MEM_LEN);

			for(i = 0; i < MEM_LEN; i++)
				par->screen[i*2] = par->buffer[i];

			for(i = 0; i < MEM_LEN; i += BLOCK_LEN) {
				/* write some stuff */	
				spi_command(par->spi, RAMWR);
				spi_write(par->spi, &par->screen[i*2], BLOCK_LEN * 2);
				spi_command(par->spi, NOP);
//			spi_data(par->spi, &par->msg);
			}
		}
		schedule_timeout_interruptible(REFR_TIME);
	}

	return 0;
}

static int __devinit pcf8833_probe(struct spi_device *spi)
{
	struct fb_info *info;
	int retval;
	struct pcf8833_par *par;
//	struct spi_message *m;
//	struct spi_transfer *x;

	spi->master->bus_num = 1;
	spi->chip_select = 0;
	spi->max_speed_hz = 8 * 1000 * 1000;
	spi->mode = SPI_MODE_0;
	spi->bits_per_word = 9;

	retval = spi_setup(spi);
	if(retval < 0)
		return retval;	

	/* Software Reset LCD */
 	spi_command(spi, SWRESET);
	mdelay(10);

	spi_command(spi, SLEEPOUT);
	mdelay(10);

	/* RGB 8 bit */
	spi_command(spi, COLMOD);
	spi_data(spi, 0x02);

	/* MY MX V LAO RGB X X X */
	spi_command(spi, MADCTL);
	spi_data(spi, 0x00);
	/* Write contrast */
	spi_command(spi, SETCON);
	spi_data(spi, 0x3f);
	mdelay(10);

	spi_command(spi, DISPON);

	retval = -ENOMEM;
		
	info = framebuffer_alloc(sizeof(struct pcf8833_par), &spi->dev);
	if(!info)
		return retval;

	info->screen_base = alloc_pages_exact(MEM_LEN, GFP_DMA | __GFP_ZERO);
	if(!info->screen_base)
		goto err;

	info->fbops = &pcf8833_ops;
	info->var = pcf8833_var;
	info->fix = pcf8833_fix;

	info->fix.smem_len = MEM_LEN;
	info->fix.smem_start = (unsigned long)virt_to_phys(info->screen_base);
	info->screen_size = info->fix.smem_len;
	info->flags = FBINFO_FLAG_DEFAULT;

	par = info->par;
	par->spi = spi;
	par->buffer = info->screen_base;

	par->shadow = kmalloc(MEM_LEN, GFP_KERNEL | GFP_DMA);
	if(!par->shadow)
		goto err1;

	par->screen = kmalloc(MEM_LEN * 2, GFP_KERNEL | GFP_DMA);
	if(!par->screen)
		goto err2;
	memset(par->screen, 0xff, MEM_LEN * 2);

	/* column start/end ram (x) */
	spi_command(spi, CASET);
	spi_data(spi, 2);            
	spi_data(spi, 129);          
	/* page start/end ram (y) */
	spi_command(spi, PASET);
	spi_data(spi, 2);               
	spi_data(spi, 129); 

//	m = &par->msg;                                                          
//	x = &par->xfer;                                       

//	spi_message_init(m);

//	x->tx_buf = par->screen;
//	x->len = MEM_LEN * 2;
//	spi_message_add_tail(x, m);

//	m->complete = spi_complete;
//	m->context = par;

	retval = register_framebuffer(info);
	if (retval < 0)
		goto err3;

	dev_set_drvdata(&spi->dev, info);

	par->pcf8833_thread_task = kthread_run(pcf8833_thread, par, "pcf8833");

	retval = -EIO;
	if(IS_ERR(par->pcf8833_thread_task))
		goto err3;
	
	printk(KERN_INFO "fb%d: %s frame buffer device, %lx base address %dK of video memory\n",
	       info->node, info->fix.id, info->fix.smem_start, info->fix.smem_len >> 10);

	return 0;
err3:
	kfree(par->screen);
err2:
	kfree(par->shadow);
err1:
	free_pages_exact(par->buffer, MEM_LEN);
err:
	framebuffer_release(info);

	return retval;
}

static int __devexit pcf8833_remove(struct spi_device *spi)
{
	struct fb_info *info = dev_get_drvdata(&spi->dev);

	if (info) {
		struct pcf8833_par *par = info->par;

		unregister_framebuffer(info);
		free_pages_exact(par->buffer, MEM_LEN);
		kfree(par->shadow);
		kfree(par->screen);
		framebuffer_release(info);

		kthread_stop(par->pcf8833_thread_task);
	}
	return 0;
}

static struct spi_driver pcf8833_driver = {
	.driver = {
		.name       = "pcf8833",
		.bus        = &spi_bus_type,
		.owner      = THIS_MODULE,
	},
	.probe      = pcf8833_probe,
	.remove     = __devexit_p(pcf8833_remove),
};

static int __init pcf8833_init(void)
{
	printk("pcf8833 spi fb driver\n");
	return spi_register_driver(&pcf8833_driver);
}

static void __exit pcf8833_exit(void)
{
	spi_unregister_driver(&pcf8833_driver);
}

module_init(pcf8833_init);
module_exit(pcf8833_exit);

MODULE_AUTHOR("Alexander Kudjashev");
MODULE_DESCRIPTION("pcf8833 spi fb driver");
MODULE_LICENSE("GPL");
